# Copyright 2018-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.template.embeddings` module.
Integration tests should be placed into ``test_templates.py``.
"""
# pylint: disable=protected-access,cell-var-from-loop
import pytest
from math import pi
import numpy as np
import pennylane as qml
import pennylane._queuing
from pennylane.templates.embeddings import (AngleEmbedding,
                                            BasisEmbedding,
                                            AmplitudeEmbedding,
                                            QAOAEmbedding,
                                            DisplacementEmbedding,
                                            SqueezingEmbedding)
from pennylane import Beamsplitter
from pennylane.wires import Wires


pytestmark = pytest.mark.usefixtures("tape_mode")


class TestAmplitudeEmbedding:
    """ Tests the AmplitudeEmbedding method."""

    FEATURES = [np.array([0, 1, 0, 0]),
                1 / np.sqrt(4) * np.array([1, 1, 1, 1]),
                np.array([np.complex(-np.sqrt(0.1), 0.0), np.sqrt(0.3),
                      np.complex(0, -np.sqrt(0.1)), np.sqrt(0.5)])]

    NOT_ENOUGH_FEATURES = [np.array([0, 1, 0]),
                           1 / np.sqrt(3) * np.array([1, 1, 1]),
                           np.array([np.complex(-np.sqrt(0.1), 0.0), np.sqrt(0.3),
                                     np.complex(0, -np.sqrt(0.6))])]

    TOO_MANY_FEATURES = [[0, 0, 0, 1, 0],
                         1 / np.sqrt(8) * np.array([1] * 8),
                         [np.complex(-np.sqrt(0.1), 0.0), np.sqrt(0.3),
                          np.complex(0, -np.sqrt(0.6)), 0., 0.]]

    @pytest.mark.parametrize("inpt", FEATURES)
    def test_prepares_correct_state(self, inpt):
        """Checks the state produced by AmplitudeEmbedding() for real and complex
        inputs."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), normalize=False)
            return [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]

        circuit(x=inpt)
        state = circuit.device.state.ravel()
        assert np.allclose(state, inpt)

    @pytest.mark.parametrize("inpt", NOT_ENOUGH_FEATURES)
    @pytest.mark.parametrize("pad", [complex(0.1, 0.1), 0., 1.])
    def test_prepares_padded_state(self, inpt, pad):
        """Checks the state produced by AmplitudeEmbedding() for real and complex padding constants."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), pad_with=pad, normalize=False)
            return [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]

        circuit(x=inpt)
        state = circuit.device.state.ravel()
        assert len(set(state[len(inpt):])) == 1

    @pytest.mark.parametrize("inpt", FEATURES)
    def test_throws_exception_if_not_normalized(self, inpt):
        """Checks that AmplitudeEmbedding() throws exception when state is not normalized and `normalize=False`."""
        not_nrmlzd = 2 * inpt
        n_qubits = 2
        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), pad_with=None, normalize=False)
            return [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]

        with pytest.raises(ValueError, match="Features must be a vector of length"):
            circuit(x=not_nrmlzd)

    def test_throws_exception_if_features_wrong_shape(self):
        """Verifies that AmplitudeEmbedding throws exception
        if features has more than one dimensions."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits))
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Features must be a one-dimensional (tensor|vector)"):
            circuit(x=[[1., 0.], [0., 0.]])

    @pytest.mark.parametrize("inpt", NOT_ENOUGH_FEATURES)
    def test_throws_exception_if_fewer_features_than_amplitudes(self, inpt):
        """Verifies that AmplitudeEmbedding() throws exception
        if the number of features is fewer than the number of amplitudes, and
        no automatic padding is chosen."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), pad_with=None, normalize=False)
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Features must be of length"):
            circuit(x=inpt)

    @pytest.mark.parametrize("inpt", TOO_MANY_FEATURES)
    def test_throws_exception_if_more_features_than_amplitudes(self, inpt):
        """Verifies that AmplitudeEmbedding() throws exception
        if the number of features is larger than the number of amplitudes, and
        no automatic padding is chosen."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), pad_with=None, normalize=False)
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Features must be of length"):
            circuit(x=inpt)

    @pytest.mark.parametrize("inpt", TOO_MANY_FEATURES)
    def test_throws_exception_if_more_features_than_amplitudes_padding(self, inpt):
        """Verifies that AmplitudeEmbedding() throws exception
        if the number of features is larger than the number of amplitudes, and
        automatic padding is chosen."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(features=x, wires=range(n_qubits), pad_with=0., normalize=False)
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Features must be of length"):
            circuit(x=inpt)

    def test_amplitude_embedding_tolerance_value(self):
        """Tests that a small enough tolerance value is used for Amplitude
        Embedding."""
        inputs = np.array([0.25895178024895, 0.115997030111517, 0.175840500169049, 0.16545033015906,
                            0.016337370015706, 0.006616800006361, 0.22326375021464, 0.161815530155566,
                            0.234776190225708, 0.082623190079432, 0.291982110280705, 0.295344560283937,
                            0.05998731005767, 0.056911140054713, 0.274260680263668, 0.163596590157278,
                            0.048460970046589, 0.292306260281016, 0.292451040281155, 0.007849840007547,
                            0.218302930209871, 0.326763300314142, 0.163634550157314, 0.275472160264832,
                            0.105510810101436])

        tolerance = 10e-10
        num_qubits = 5
        dev = qml.device('default.qubit', wires=num_qubits)
        assert np.isclose(np.sum(np.abs(inputs) ** 2), 1, tolerance)

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(x, list(range(num_qubits)), pad_with=0., normalize=True)
            return qml.expval(qml.PauliZ(0))

        # No normalization error is raised
        circuit(x=inputs)

    def test_deprecated_pad_arg(self):
        """Test that the pad argument raises a deprecation warning"""

        num_qubits = 2
        dev = qml.device('default.qubit', wires=num_qubits)
        inputs = np.array([1.0, 0.0, 0.0, 0.0])

        @qml.qnode(dev)
        def circuit(x=None):
            AmplitudeEmbedding(x, list(range(num_qubits)), pad=0., normalize=True)
            return qml.expval(qml.PauliZ(0))

        with pytest.warns(PendingDeprecationWarning, match="will be replaced by the pad_with option in future versions"):
            circuit(x=inputs)


class TestAngleEmbedding:
    """ Tests the AngleEmbedding method."""

    def test_state_rotx(self, qubit_device, n_subsystems):
        """Checks the state produced using the rotation='X' strategy."""

        features = [pi / 2, pi / 2, pi / 4, 0]

        @qml.qnode(qubit_device)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='X')
            qml.PauliX(wires=0)
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='X')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        target = [1, -1, 0, 1, 1]

        assert np.allclose(res, target[:n_subsystems])

    def test_state_roty(self, qubit_device, n_subsystems):
        """Checks the state produced using the rotation='Y' strategy."""

        features = [pi / 2, pi / 2, pi / 4, 0]

        @qml.qnode(qubit_device)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Y')
            qml.PauliX(wires=0)
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Y')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        target = [-1, -1, 0, 1, 1]
        assert np.allclose(res, target[:n_subsystems])

    def test_state_rotz(self, qubit_device, n_subsystems):
        """Checks the state using the rotation='Z' strategy."""

        features = [pi / 2, pi / 2, pi / 4, 0]

        @qml.qnode(qubit_device)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Z')
            qml.PauliX(wires=0)
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Z')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        target = [-1, 1, 1, 1, 1]
        assert np.allclose(res, target[:n_subsystems])

    @pytest.mark.parametrize('strategy', ['X', 'Y', 'Z'])
    def test_angle_embedding_fewer_features(self, strategy):
        """Tests case with fewer features than rotation gates."""
        features = [pi / 2, pi / 2, pi / 4, 0]
        n_subsystems = 5
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Z')
            qml.PauliX(wires=0)
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='Z')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features)
        target = [-1, 1, 1, 1, 1]
        assert np.allclose(res, target)

    @pytest.mark.parametrize('strategy', ['X', 'Y', 'Z'])
    def test_exception_fewer_rotations(self, strategy):
        """Verifies that exception is raised if there are fewer
           rotation gates than features."""

        features = [0, 0, 0, 0]
        n_subsystems = 1
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation=strategy)
            qml.PauliX(wires=0)
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation=strategy)
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        with pytest.raises(ValueError, match="Features must be of"):
            circuit(x=features)

    def test_exception_wrongrot(self):
        """Verifies that exception is raised if the
        rotation strategy is unknown."""

        n_subsystems = 1
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='A')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        with pytest.raises(ValueError, match="Rotation option"):
            circuit(x=[1])

    def test_exception_wrong_dim(self):
        """Verifies that exception is raised if the
        number of dimensions of features is incorrect."""
        if not qml.tape_mode_active():
            pytest.skip("This validation is only performed in tape mode")

        n_subsystems = 1
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            AngleEmbedding(features=x, wires=range(n_subsystems), rotation='A')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        with pytest.raises(ValueError, match="Features must be a one-dimensional"):
            circuit(x=[[1], [0]])


class TestBasisEmbedding:
    """ Tests the BasisEmbedding method."""

    def test_state(self):
        """Checks the state."""

        state = np.array([0, 1])
        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            BasisEmbedding(features=x, wires=range(2))
            return [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]

        res = circuit(x=state)
        assert np.allclose(res, [1, -1])

    def test_too_many_input_bits_exception(self):
        """Verifies that exception thrown if there are more features than qubits."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            BasisEmbedding(features=x, wires=range(2))
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError):
            circuit(x=np.array([0, 1, 1]))

    def test_not_enough_input_bits_exception(self):
        """Verifies that exception thrown if there are less features than qubits."""

        n_qubits = 2
        dev = qml.device('default.qubit', wires=n_qubits)

        @qml.qnode(dev)
        def circuit(x=None):
            BasisEmbedding(features=x, wires=range(2))
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError):
            circuit(x=np.array([0]))

    def test_input_not_binary_exception(self):
        """Verifies that exception raised if the features contain
        values other than zero and one."""

        n_subsystems = 2
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            BasisEmbedding(features=x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Basis state must only consist of"):
            circuit(x=[2, 3])

    def test_exception_wrong_dim(self):
        """Verifies that exception is raised if the
        number of dimensions of features is incorrect."""
        if not qml.tape_mode_active():
            pytest.skip("This validation is only performed in tape mode")

        n_subsystems = 2
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            BasisEmbedding(features=x, wires=[0, 1])
            return qml.expval(qml.PauliZ(0))

        with pytest.raises(ValueError, match="Features must be one-dimensional"):
            circuit(x=[[1], [0]])


class TestIQPEmbedding:
    """ Tests the IQPEmbedding method."""

    QUEUES = [(0, []),
              (1, [qml.Hadamard, qml.RZ]),
              (2, [qml.Hadamard, qml.Hadamard, qml.RZ, qml.RZ, qml.MultiRZ]),
              (3, [qml.Hadamard, qml.Hadamard, qml.Hadamard, qml.RZ, qml.RZ, qml.RZ,
                   qml.MultiRZ, qml.MultiRZ, qml.MultiRZ])]

    @pytest.mark.parametrize('n_wires, expected_queue', QUEUES)
    @pytest.mark.parametrize('n_repeats', [1, 2])
    def test_queue_default_pattern(self, n_wires, expected_queue, n_repeats):
        """Checks the queue for the default pattern."""

        with pennylane._queuing.OperationRecorder() as rec:
            qml.templates.IQPEmbedding(features=list(range(n_wires)), wires=range(n_wires), n_repeats=n_repeats)

        expected_queue = expected_queue * n_repeats

        for gate, expected_gate in zip(rec.queue, expected_queue):
            assert isinstance(gate, expected_gate)

    @pytest.mark.parametrize('features, expected_params', [([1., 2., 3.],
                                                            [1., 2., 3., 1 * 2, 1 * 3, 2 * 3]),
                                                           ([0.1, 0.2, 0.3],
                                                            [0.1, 0.2, 0.3, 0.1 * 0.2, 0.1 * 0.3, 0.2 * 0.3])])
    @pytest.mark.parametrize('wires', [range(3),
                                       [2, 0, 1]])
    def test_queue_parameters(self, features, expected_params, wires):
        """Checks the queued parameters, for consecutive and non-consecutive ``wires`` argument."""

        with pennylane._queuing.OperationRecorder() as rec:
            qml.templates.IQPEmbedding(features=features, wires=wires)

        # compare all nonempty gate parameters to expected ones
        counter = 0
        for gate in rec.queue:
            if gate.parameters:
                assert gate.parameters[0] == expected_params[counter]
                counter += 1

    @pytest.mark.parametrize('wires, expected_queue_wires', [(range(3), [[0], [1], [2], [0], [1], [2],
                                                                         [0, 1], [0, 2], [1, 2]]),
                                                             ([2, 0, 1], [[2], [0], [1], [2], [0], [1],
                                                                          [2, 0], [2, 1], [0, 1]])])
    def test_queue_correct_wires(self, wires, expected_queue_wires):
        """Checks the queued wires for a consecutive and non-consecutive sequence
           of indices in the ``wires`` argument."""

        with pennylane._queuing.OperationRecorder() as rec:
            qml.templates.IQPEmbedding(features=list(range(3)), wires=wires)

        # compare all gate wires to expected ones
        for idx, gate in enumerate(rec.queue):
            assert gate.wires == Wires(expected_queue_wires[idx])

    @pytest.mark.parametrize('pattern', [[[0, 3], [1, 2], [2, 0]],
                                         [[2, 3], [0, 2], [1, 0]]])
    def test_wires_custom_pattern(self, pattern):
        """Checks the queue for a custom pattern."""

        with pennylane._queuing.OperationRecorder() as rec:
            qml.templates.IQPEmbedding(features=list(range(4)), wires=range(4), pattern=pattern)

        counter = 0
        for gate in rec.queue:
            # check wires of entanglers
            if len(gate.wires) == 2:
                assert gate.wires == Wires(pattern[counter])
                counter += 1

    @pytest.mark.parametrize('features', [[1., 2.],
                                          [1., 2., 3., 4.],
                                          [[1., 1.], [2., 2.], [3., 3.]]])
    def test_exception_wrong_number_of_features(self, features):
        """Verifies that an exception is raised if 'feature' has the wrong shape."""

        dev = qml.device('default.qubit', wires=3)

        @qml.qnode(dev)
        def circuit(f=None):
            qml.templates.IQPEmbedding(features=f, wires=range(3))
            return [qml.expval(qml.PauliZ(w)) for w in range(3)]

        with pytest.raises(ValueError, match="Features must be"):
            circuit(f=features)

    def test_exception_incorrect_pattern(self):
        """Verifies that an exception is raised if 'pattern' has the wrong shape."""
        if qml.tape_mode_active():
            pytest.skip("Check only done in non-tape mode")

        dev = qml.device('default.qubit', wires=3)
        features = [1., 2., 3.]

        @qml.qnode(dev)
        def circuit(f=None):
            qml.templates.IQPEmbedding(features=f, wires=range(3), pattern=[0., 0.2])
            return [qml.expval(qml.PauliZ(w)) for w in range(3)]

        with pytest.raises(ValueError, match="'pattern' must be a"):
            circuit(f=features)


class TestQAOAEmbedding:
    """ Tests the QAOAEmbedding method."""

    QUEUES = [(1, (1, 1), [qml.RX, qml.RY, qml.RX]),
              (2, (1, 3), [qml.RX, qml.RX, qml.MultiRZ, qml.RY, qml.RY, qml.RX, qml.RX]),
              (3, (1, 6), [qml.RX, qml.RX, qml.RX, qml.MultiRZ, qml.MultiRZ, qml.MultiRZ,
                   qml.RY, qml.RY, qml.RY, qml.RX, qml.RX, qml.RX])]

    @pytest.mark.parametrize('n_wires, weight_shape, expected_queue', QUEUES)
    def test_queue(self, n_wires, weight_shape, expected_queue):
        """Checks the queue for the default settings."""

        with pennylane._queuing.OperationRecorder() as rec:
            QAOAEmbedding(features=list(range(n_wires)), weights=np.zeros(shape=weight_shape), wires=range(n_wires))

        for gate, expected_gate in zip(rec.queue, expected_queue):
            assert isinstance(gate, expected_gate)

    def test_state_zero_weights(self, qubit_device, n_subsystems, tol):
        """Checks the state produced by QAOAEmbedding() is correct if the weights are zero."""

        features = [pi, pi / 2, pi / 4, 0]
        if n_subsystems == 1:
            shp = (1, 1)
        elif n_subsystems == 2:
            shp = (1, 3)
        else:
            shp = (1, 2 * n_subsystems)

        weights = np.zeros(shape=shp)

        @qml.qnode(qubit_device)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_subsystems))
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        target = [1, -1, 0, 1, 1]
        assert np.allclose(res, target[:n_subsystems], atol=tol, rtol=0)

    @pytest.mark.parametrize('n_subsystems, weights, target', [(1, [[pi / 2]], [0]),
                                                               (2, [[1, pi / 2, pi / 4]], [0, 1 / np.sqrt(2)]),
                                                               (3, [[0, 0, 0, pi, pi / 2, pi / 4]],
                                                                [-1, 0, 1 / np.sqrt(2)])])
    def test_output_local_field_ry(self, n_subsystems, weights, target, tol):
        """Checks the output if the features are zero. Uses RY local fields."""

        features = np.zeros(shape=(n_subsystems,))
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_subsystems), local_field='Y')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        assert np.allclose(res, target, atol=tol, rtol=0)

    @pytest.mark.parametrize('n_subsystems, weights, target', [(1, [[pi / 2]], [0]),
                                                               (2, [[1, pi / 2, pi / 4]], [0, 1 / np.sqrt(2)]),
                                                               (3, [[0, 0, 0, pi, pi / 2, pi / 4]],
                                                                [-1, 0, 1 / np.sqrt(2)])])
    def test_output_local_field_rx(self, n_subsystems, weights, target, tol):
        """Checks the output if the features are zero. Uses RX local fields."""

        features = np.zeros(shape=(n_subsystems,))
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_subsystems), local_field='X')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        assert np.allclose(res, target, atol=tol, rtol=0)

    @pytest.mark.parametrize('n_subsystems, weights, target', [(1, [[pi / 2]], [1]),
                                                               (2, [[1, pi / 2, pi / 4]], [1, 1]),
                                                               (3, [[0, 0, 0, pi, pi / 2, pi / 4]], [1, 1, 1])])
    def test_output_local_field_rz(self, n_subsystems, weights, target, tol):
        """Checks the output if the features are zero. Uses RZ local fields."""

        features = np.zeros(shape=(n_subsystems,))
        dev = qml.device('default.qubit', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_subsystems), local_field='Z')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_subsystems)]

        res = circuit(x=features[:n_subsystems])
        assert np.allclose(res, target, atol=tol, rtol=0)

    @pytest.mark.parametrize('weights, target', [([[np.pi, 0, 0]], [1, 1]),
                                                 ([[np.pi / 2, 0, 0]], [0, 0]),
                                                 ([[0, 0, 0]], [-1, -1])])
    def test_output_zz(self, weights, target, tol):
        """Checks the output if the features and entangler weights are nonzero."""

        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(2))
            return [qml.expval(qml.PauliZ(i)) for i in range(2)]

        res = circuit(x=[np.pi/2, np.pi/2])

        assert np.allclose(res, target, atol=tol, rtol=0)

    @pytest.mark.parametrize('n_wires, features, weights, target', [(2, [0], [[0, 0, np.pi / 2]], [1, 0]),
                                                                    (3, [0, 0], [[0, 0, 0, 0, 0, np.pi / 2]],
                                                                     [1, 1, 0])])
    def test_state_more_qubits_than_features(self, n_wires, features, weights, target, tol):
        """Checks the state is correct if there are more qubits than features."""

        dev = qml.device('default.qubit', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_wires), local_field='Z')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_wires)]

        res = circuit(x=features)
        assert np.allclose(res, target, atol=tol, rtol=0)

    def test_exception_fewer_wires_than_features(self, ):
        """Verifies that exception raised if there are fewer
           wires than features."""

        features = [0, 0, 0, 0]
        n_wires = 1
        weights = np.zeros(shape=(1, 2 * n_wires))
        dev = qml.device('default.qubit', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_wires))
            return [qml.expval(qml.PauliZ(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="Features must be of "):
            circuit(x=features)

    def test_exception_wrongrot(self):
        """Verifies exception raised if the
        rotation strategy is unknown."""

        n_wires = 1
        weights = np.zeros(shape=(1, 1))
        dev = qml.device('default.qubit', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_wires), local_field='A')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="did not recognize"):
            circuit(x=[1])

    def test_exception_wrong_dim(self):
        """Verifies that exception is raised if the
        number of dimensions of features is incorrect."""
        if not qml.tape_mode_active():
            pytest.skip("This validation is only performed in tape mode")

        n_wires = 1
        weights = np.zeros(shape=(1, 1))
        dev = qml.device('default.qubit', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            QAOAEmbedding(features=x, weights=weights, wires=range(n_wires), local_field='A')
            return [qml.expval(qml.PauliZ(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="Features must be a one-dimensional"):
            circuit(x=[[1], [0]])


class TestDisplacementEmbedding:
    """ Tests the DisplacementEmbedding method."""

    def test_state_execution_amplitude(self):
        """Checks the state using the amplitude execution method."""

        features = np.array([0.1, 1.2])
        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            DisplacementEmbedding(features=x, wires=range(n_wires), method='amplitude', c=1.)
            return [qml.expval(qml.NumberOperator(wires=0)), qml.expval(qml.NumberOperator(wires=1))]

        assert np.allclose(circuit(x=features), [0.01, 1.44], atol=0.001)

    def test_state_execution_phase(self):
        """Checks the state using the phase execution method."""

        features = np.array([1.2, 0.3])
        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            DisplacementEmbedding(features=x, wires=range(n_wires), method='phase', c=1.)
            Beamsplitter(pi / 2, 0, wires=[0, 1])
            DisplacementEmbedding(features=[0, 0], wires=range(n_wires), method='phase', c=1.)
            return [qml.expval(qml.NumberOperator(wires=0)), qml.expval(qml.NumberOperator(wires=1))]

        assert np.allclose(circuit(x=features), [0.089327, 2.724715], atol=0.01)

    def test_exception_for_wrong_num_wires(self):
        """Verifies that an exception is thrown if number of subsystems wrong."""

        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            DisplacementEmbedding(features=x, wires=range(n_wires), method='phase')
            return [qml.expval(qml.X(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="Features must be of"):
            circuit(x=[0.2, 0.3, 0.4])

    def test_strategy_not_recognized_exception(self):
        """Verifies that an exception is thrown if is strategy unknown."""

        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            DisplacementEmbedding(features=x, wires=range(n_wires), method='A')
            return [qml.expval(qml.X(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="did not recognize"):
            circuit(x=[1, 2])

    def test_exception_wrong_dim(self):
        """Verifies that exception is raised if the
        number of dimensions of features is incorrect."""
        if not qml.tape_mode_active():
            pytest.skip("This validation is only performed in tape mode")

        n_subsystems = 2
        dev = qml.device('default.gaussian', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            DisplacementEmbedding(features=x, wires=[0, 1])
            return qml.expval(qml.X(0))

        with pytest.raises(ValueError, match="Features must be a one-dimensional"):
            circuit(x=[[1], [0]])


class TestSqueezingEmbedding:
    """ Tests the SqueezingEmbedding method."""

    def test_state_execution_amplitude(self):
        """Checks the state using the amplitude execution method."""

        features = np.array([1.2, 0.3])
        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            SqueezingEmbedding(features=x, wires=range(n_wires), method='amplitude', c=1)
            return [qml.expval(qml.NumberOperator(wires=0)), qml.expval(qml.NumberOperator(wires=1))]

        assert np.allclose(circuit(x=features), [2.2784, 0.09273], atol=0.001)

    def test_state_execution_phase(self):
        """Checks the state produced using the phase execution method."""

        features = np.array([1.2, 0.3])
        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            SqueezingEmbedding(features=x, wires=range(n_wires), method='phase', c=1)
            Beamsplitter(pi / 2, 0, wires=[0, 1])
            SqueezingEmbedding(features=[0, 0], wires=range(n_wires), method='phase', c=1)
            return [qml.expval(qml.NumberOperator(wires=0)), qml.expval(qml.NumberOperator(wires=1))]

        assert np.allclose(circuit(x=features), [12.86036, 8.960306], atol=0.001)

    def test_exception_for_wrong_num_wires(self):
        """Verifies that an exception is thrown if number of modes is wrong."""

        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            SqueezingEmbedding(features=x, wires=range(n_wires), method='phase')
            return [qml.expval(qml.X(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="Features must be of "):
            circuit(x=[0.2, 0.3, 0.4])

    def test_strategy_not_recognized_exception(self):
        """Verifies that an exception is thrown if strategy unknown."""

        n_wires = 2
        dev = qml.device('default.gaussian', wires=n_wires)

        @qml.qnode(dev)
        def circuit(x=None):
            SqueezingEmbedding(features=x, wires=range(n_wires), method='A')
            return [qml.expval(qml.X(i)) for i in range(n_wires)]

        with pytest.raises(ValueError, match="did not recognize"):
            circuit(x=[1, 2])

    def test_exception_wrong_dim(self):
        """Verifies that exception is raised if the
        number of dimensions of features is incorrect."""
        if not qml.tape_mode_active():
            pytest.skip("This validation is only performed in tape mode")

        n_subsystems = 2
        dev = qml.device('default.gaussian', wires=n_subsystems)

        @qml.qnode(dev)
        def circuit(x=None):
            SqueezingEmbedding(features=x, wires=[0, 1])
            return qml.expval(qml.X(0))

        with pytest.raises(ValueError, match="Features must be one-dimensional"):
            circuit(x=[[1], [0]])
